module BinarySeachTreeTests where

import qualified BinarySearchTree as BST

tests :: IO ()
tests = let
    testTrue :: String -> String -> Bool -> IO()
    testTrue name message result = putStrLn $ name ++ ": " ++
        (if result then "OK" else "FAILED (" ++ message ++ ")")
    testEquals :: (Eq a, Show a) => String -> String -> a -> a -> IO()
    testEquals name message left right = putStrLn $ name ++ ": " ++
        (if left == right then "OK" else "FAILED (" ++ message ++ ": " ++
        show left ++ " /= " ++ show right ++ ")")
    testNotEquals :: (Eq a, Show a) => String -> String -> a -> a -> IO()
    testNotEquals name message left right = putStrLn $ name ++ ": " ++
        (if left /= right then "OK" else "FAILED (" ++ message ++ ": " ++
        show left ++ " == " ++ show right ++ ")")
    insertMultiple :: Ord a => [a] -> BST.BinaryTree a
    insertMultiple = foldr BST.insert BST.empty
    exampleTree :: BST.BinaryTree Int
    exampleTree = insertMultiple [13, 7, 12, 2, 16, 19, 6, 14]
    in do
        testNotEquals "Insert to empty" ("Inserting into empty tree results" ++
            " in empty tree") BST.empty (BST.insert 4 BST.empty)
        testEquals "Insert duplicate" ("Inserting value that is already in" ++
            " the tree changes it") (insertMultiple [1..5])
            (insertMultiple (5 : [1..5]))
        testNotEquals "Insert order" ("Inserting 1..5 in different order" ++
            " results in the same tree") (insertMultiple [1..5])
            (insertMultiple [5,4..1])
        testEquals "Insert example tree" ("Creating example tree in different" ++
            " ways results in different trees")
            (insertMultiple [6, 7, 13, 7, 12, 2, 16, 19, 6, 14])
            (insertMultiple [13, 2, 19, 7, 16, 12, 6, 19, 14, 14])
        testEquals "Fold empty tree" ("Folding an empty tree does not result" ++
            " in base value") 3 (BST.foldTree 3 (\ _ _ _ -> 0) BST.empty)
        testEquals "Fold example tree by adding" ("Sum of values in example" ++
            " tree is not 89") 89
            (BST.foldTree 0 (\ l x r -> l + x + r) exampleTree)
        testEquals "Fold with non commutative" ("FoldTree gives wrong order of" ++
            " arguments to its function") (-5)
            (BST.foldTree 0 (\ l x r -> x - l + r) exampleTree)
        testEquals "Height of empty tree" "Empty tree does not have height -1"
            (-1) (BST.height BST.empty)
        testEquals "Height of root only" "Single node tree doesn't have height 0"
            0 (BST.height $ BST.insert 5 BST.empty)
        testEquals "Height of mirrored tree" "Mirrored trees have different height"
            (BST.height $ insertMultiple [1..5])
            (BST.height $ insertMultiple [5,4..1])
        testEquals "Height of example tree" "Example tree does not have height 3"
            3 (BST.height exampleTree)
        testTrue "Is not member of empty tree" "isMember on empty tree is true"
            (not $ any (`BST.isMember` BST.empty) [1..10])
        testTrue "Is member after inserting" ("isMember returns false for some" ++
            " elements that where inserted")
            (all (`BST.isMember` insertMultiple [1..5]) [1..5])
        testTrue "Is not member after inserting different" ("isMember returns" ++
            " true for elements that where not inserted")
            (not $ any (`BST.isMember` insertMultiple [1..5]) [6..10])
        testTrue "Is member in example tree" "Members are missing in example tree"
            (all (`BST.isMember` exampleTree) [2, 6, 7, 12, 13, 14, 16, 19])
